"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BytecodeExpressionGenerator = void 0;
const greybel_core_1 = require("greybel-core");
const miniscript_core_1 = require("miniscript-core");
const path_1 = require("path");
const default_1 = require("../types/default");
const number_1 = require("../types/number");
const string_1 = require("../types/string");
const error_1 = require("../utils/error");
const instruction_1 = require("./instruction");
const keywords_1 = require("./keywords");
const utils_1 = require("./utils");
class BytecodeExpressionGenerator {
    constructor(context, parseCodeFunction, stmtGenerator) {
        this.context = context;
        this.parseCode = parseCodeFunction;
        this.stmtGenerator = stmtGenerator;
    }
    process(node, context) {
        return __awaiter(this, void 0, void 0, function* () {
            switch (node.type) {
                case miniscript_core_1.ASTType.MemberExpression:
                    yield this.processMemberExpression(node, context);
                    return;
                case miniscript_core_1.ASTType.IndexExpression:
                    yield this.processIndexExpression(node);
                    return;
                case miniscript_core_1.ASTType.SliceExpression:
                    yield this.processSliceExpression(node);
                    return;
                case miniscript_core_1.ASTType.Identifier:
                    yield this.processIdentifier(node);
                    return;
                case miniscript_core_1.ASTType.BooleanLiteral:
                case miniscript_core_1.ASTType.StringLiteral:
                case miniscript_core_1.ASTType.NumericLiteral:
                case miniscript_core_1.ASTType.NilLiteral:
                    yield this.processLiteral(node);
                    return;
                case miniscript_core_1.ASTType.IsaExpression:
                case miniscript_core_1.ASTType.BinaryExpression:
                case miniscript_core_1.ASTType.LogicalExpression:
                    yield this.processEvaluationExpression(node);
                    return;
                case miniscript_core_1.ASTType.MapConstructorExpression:
                    yield this.processMapConstructorExpression(node);
                    return;
                case miniscript_core_1.ASTType.ListConstructorExpression:
                    yield this.processListConstructorExpression(node);
                    return;
                case miniscript_core_1.ASTType.FunctionDeclaration:
                    yield this.processFunctionDeclaration(node, context);
                    return;
                case miniscript_core_1.ASTType.ParenthesisExpression:
                    yield this.process(node.expression);
                    return;
                case miniscript_core_1.ASTType.BinaryNegatedExpression:
                case miniscript_core_1.ASTType.UnaryExpression:
                case miniscript_core_1.ASTType.NegationExpression:
                    yield this.processUnaryExpression(node);
                    return;
                case miniscript_core_1.ASTType.CallExpression:
                    yield this.processCallExpression(node);
                    return;
                case miniscript_core_1.ASTType.EmptyExpression:
                    this.context.pushCode({
                        op: instruction_1.OpCode.PUSH,
                        value: default_1.DefaultType.Void
                    }, node);
                    return;
                case greybel_core_1.ASTType.FeatureEnvarExpression:
                    yield this.processEnvarExpression(node);
                    return;
                case greybel_core_1.ASTType.FeatureLineExpression:
                    this.context.pushCode({
                        op: instruction_1.OpCode.PUSH,
                        value: new number_1.CustomNumber(node.start.line)
                    }, node);
                    return;
                case greybel_core_1.ASTType.FeatureFileExpression:
                    this.context.pushCode({
                        op: instruction_1.OpCode.PUSH,
                        value: new string_1.CustomString((0, path_1.basename)(this.context.target.peek()))
                    }, node);
                    return;
                case miniscript_core_1.ASTType.Comment:
                    return;
                default: {
                    const range = new miniscript_core_1.ASTRange(node.start, node.end);
                    throw new error_1.PrepareError(`Unexpected AST type ${node.type}`, {
                        target: this.context.target.peek(),
                        range
                    });
                }
            }
        });
    }
    processMemberExpression(node, context) {
        return __awaiter(this, void 0, void 0, function* () {
            const base = (0, utils_1.unwrap)(node.base);
            if (base instanceof miniscript_core_1.ASTIdentifier && base.name === keywords_1.RuntimeKeyword.Super) {
                this.context.pushCode({
                    op: instruction_1.OpCode.PUSH,
                    value: new string_1.CustomString(node.identifier.name)
                }, node.identifier);
                this.context.pushCode({
                    op: instruction_1.OpCode.GET_SUPER_PROPERTY,
                    invoke: !(context === null || context === void 0 ? void 0 : context.isReference)
                }, node.identifier);
            }
            else {
                yield this.process(base);
                yield this.processIdentifier(node.identifier, {
                    isDescending: true,
                    isReference: !!(context === null || context === void 0 ? void 0 : context.isReference)
                });
            }
        });
    }
    processIndexExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const base = (0, utils_1.unwrap)(node.base);
            if (base instanceof miniscript_core_1.ASTIdentifier && base.name === keywords_1.RuntimeKeyword.Super) {
                yield this.process(node.index);
                this.context.pushCode({
                    op: instruction_1.OpCode.GET_SUPER_PROPERTY,
                    invoke: false
                }, node.index, node.type);
            }
            else {
                yield this.process(base);
                yield this.process(node.index);
                this.context.pushCode({
                    op: instruction_1.OpCode.GET_PROPERTY,
                    invoke: false
                }, node.index, node.type);
            }
        });
    }
    processSliceExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.process(node.base);
            yield this.process(node.left);
            yield this.process(node.right);
            this.context.pushCode({
                op: instruction_1.OpCode.SLICE
            }, node);
        });
    }
    processIdentifier(node, context) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!(context === null || context === void 0 ? void 0 : context.isDescending)) {
                switch (node.name) {
                    case keywords_1.RuntimeKeyword.Self: {
                        this.context.pushCode({
                            op: instruction_1.OpCode.GET_SELF
                        }, node);
                        return;
                    }
                    case keywords_1.RuntimeKeyword.Super: {
                        this.context.pushCode({
                            op: instruction_1.OpCode.GET_SUPER
                        }, node);
                        return;
                    }
                    case keywords_1.RuntimeKeyword.Outer: {
                        this.context.pushCode({
                            op: instruction_1.OpCode.GET_OUTER
                        }, node);
                        return;
                    }
                    case keywords_1.RuntimeKeyword.Locals: {
                        this.context.pushCode({
                            op: instruction_1.OpCode.GET_LOCALS
                        }, node);
                        return;
                    }
                    case keywords_1.RuntimeKeyword.Globals: {
                        this.context.pushCode({
                            op: instruction_1.OpCode.GET_GLOBALS
                        }, node);
                        return;
                    }
                    default: {
                        this.context.pushCode({
                            op: instruction_1.OpCode.GET_VARIABLE,
                            property: new string_1.CustomString(node.name),
                            invoke: !(context === null || context === void 0 ? void 0 : context.isReference)
                        }, node);
                    }
                }
            }
            else {
                this.context.pushCode({
                    op: instruction_1.OpCode.PUSH,
                    value: new string_1.CustomString(node.name)
                }, node);
                this.context.pushCode({
                    op: instruction_1.OpCode.GET_PROPERTY,
                    invoke: !(context === null || context === void 0 ? void 0 : context.isReference)
                }, node);
            }
        });
    }
    processLiteral(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const value = (0, utils_1.generateCustomValueFromASTLiteral)(node);
            this.context.pushCode({
                op: instruction_1.OpCode.PUSH,
                value
            }, node);
        });
    }
    processEvaluationExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const skip = {
                op: instruction_1.OpCode.NOOP
            };
            yield this.process(node.left);
            if (node.operator === miniscript_core_1.Operator.And) {
                this.context.pushCode({
                    op: instruction_1.OpCode.GOTO_A_IF_FALSE_AND_PUSH,
                    goto: skip
                }, node);
            }
            else if (node.operator === miniscript_core_1.Operator.Or) {
                this.context.pushCode({
                    op: instruction_1.OpCode.GOTO_A_IF_TRUE_AND_PUSH,
                    goto: skip
                }, node);
            }
            yield this.process(node.right);
            switch (node.operator) {
                case miniscript_core_1.Operator.Isa: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.ISA
                    }, node);
                    break;
                }
                case miniscript_core_1.Operator.Equal: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.EQUAL
                    }, node);
                    break;
                }
                case miniscript_core_1.Operator.NotEqual: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.NOT_EQUAL
                    }, node);
                    break;
                }
                case miniscript_core_1.Operator.LessThan: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.LESS_THAN
                    }, node);
                    break;
                }
                case miniscript_core_1.Operator.LessThanOrEqual: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.LESS_THAN_OR_EQUAL
                    }, node);
                    break;
                }
                case miniscript_core_1.Operator.GreaterThan: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.GREATER_THAN
                    }, node);
                    break;
                }
                case miniscript_core_1.Operator.GreaterThanOrEqual: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.GREATER_THAN_OR_EQUAL
                    }, node);
                    break;
                }
                case miniscript_core_1.Operator.And: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.AND
                    }, node);
                    this.context.pushCode(skip, node);
                    break;
                }
                case miniscript_core_1.Operator.Or: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.OR
                    }, node);
                    this.context.pushCode(skip, node);
                    break;
                }
                case miniscript_core_1.Operator.Plus: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.ADD
                    }, node);
                    break;
                }
                case miniscript_core_1.Operator.Minus: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.SUB
                    }, node);
                    break;
                }
                case miniscript_core_1.Operator.Asterik: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.MUL
                    }, node);
                    break;
                }
                case miniscript_core_1.Operator.Slash: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.DIV
                    }, node);
                    break;
                }
                case miniscript_core_1.Operator.Modulo: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.MOD
                    }, node);
                    break;
                }
                case miniscript_core_1.Operator.Power: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.POW
                    }, node);
                    break;
                }
                case greybel_core_1.Operator.BitwiseAnd: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.BITWISE_AND
                    }, node);
                    break;
                }
                case greybel_core_1.Operator.BitwiseOr: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.BITWISE_OR
                    }, node);
                    break;
                }
                case greybel_core_1.Operator.LeftShift: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.BITWISE_LEFT_SHIFT
                    }, node);
                    break;
                }
                case greybel_core_1.Operator.RightShift: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.BITWISE_RIGHT_SHIFT
                    }, node);
                    break;
                }
                case greybel_core_1.Operator.UnsignedRightShift: {
                    this.context.pushCode({
                        op: instruction_1.OpCode.BITWISE_UNSIGNED_RIGHT_SHIFT
                    }, node);
                    break;
                }
                default:
                    throw new Error(`Unexpected evaluation expression operator. ("${node.operator}")`);
            }
        });
    }
    processMapConstructorExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const field of node.fields) {
                yield this.process(field.key);
                yield this.process(field.value);
            }
            this.context.pushCode({
                op: instruction_1.OpCode.CONSTRUCT_MAP,
                length: node.fields.length
            }, node);
        });
    }
    processListConstructorExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const field of node.fields) {
                yield this.process(field.value);
            }
            this.context.pushCode({
                op: instruction_1.OpCode.CONSTRUCT_LIST,
                length: node.fields.length
            }, node);
        });
    }
    processFunctionDeclaration(node, context) {
        return __awaiter(this, void 0, void 0, function* () {
            const args = [];
            for (const item of node.parameters) {
                if (item instanceof miniscript_core_1.ASTIdentifier) {
                    args.push({
                        name: new string_1.CustomString(item.name),
                        defaultValue: default_1.DefaultType.Void
                    });
                }
                else if (item instanceof miniscript_core_1.ASTAssignmentStatement) {
                    let defaultValue = default_1.DefaultType.Void;
                    if (item.init instanceof miniscript_core_1.ASTLiteral) {
                        defaultValue = (0, utils_1.generateCustomValueFromASTLiteral)(item.init);
                    }
                    args.push({
                        name: new string_1.CustomString(item.variable.name),
                        defaultValue
                    });
                }
            }
            const mod = this.context.module.peek();
            mod.pushContext();
            for (const item of node.body) {
                yield this.stmtGenerator.process(item);
            }
            this.context.pushInternalCode({
                op: instruction_1.OpCode.PUSH,
                value: default_1.DefaultType.Void
            });
            this.context.pushInternalCode({
                op: instruction_1.OpCode.RETURN
            });
            const fnCode = mod.popContext().code;
            this.context.pushCode({
                op: instruction_1.OpCode.FUNCTION_DEFINITION,
                arguments: args,
                code: fnCode,
                /*
                Can be removed after MiniScript fixed outer context bug.
              */
                ignoreOuter: !(context === null || context === void 0 ? void 0 : context.includeOuter)
            }, node);
        });
    }
    processUnaryExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const arg = (0, utils_1.unwrap)(node.argument);
            switch (node.operator) {
                case miniscript_core_1.Operator.Reference:
                    if (arg instanceof miniscript_core_1.ASTMemberExpression) {
                        yield this.processMemberExpression(arg, { isReference: true });
                    }
                    else if (arg instanceof miniscript_core_1.ASTIndexExpression) {
                        yield this.processIndexExpression(arg);
                    }
                    else if (arg instanceof miniscript_core_1.ASTIdentifier) {
                        yield this.processIdentifier(arg, {
                            isDescending: false,
                            isReference: true
                        });
                    }
                    else {
                        yield this.process(arg);
                    }
                    return;
                case miniscript_core_1.Operator.Not: {
                    yield this.process(arg);
                    this.context.pushCode({
                        op: instruction_1.OpCode.FALSIFY
                    }, node);
                    return;
                }
                case miniscript_core_1.Operator.Minus: {
                    yield this.process(arg);
                    this.context.pushCode({
                        op: instruction_1.OpCode.NEGATE
                    }, node);
                    return;
                }
                case miniscript_core_1.Operator.New: {
                    yield this.process(arg);
                    this.context.pushCode({
                        op: instruction_1.OpCode.NEW
                    }, node);
                }
            }
        });
    }
    processCallExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const pushArgs = () => __awaiter(this, void 0, void 0, function* () {
                for (const arg of node.arguments) {
                    yield this.process(arg);
                }
            });
            const left = (0, utils_1.unwrap)(node.base);
            if (left instanceof miniscript_core_1.ASTMemberExpression) {
                const base = (0, utils_1.unwrap)(left.base);
                if (base instanceof miniscript_core_1.ASTIdentifier && base.name === keywords_1.RuntimeKeyword.Super) {
                    this.context.pushCode({
                        op: instruction_1.OpCode.PUSH,
                        value: new string_1.CustomString(left.identifier.name)
                    }, left.identifier);
                    yield pushArgs();
                    this.context.pushCode({
                        op: instruction_1.OpCode.CALL_SUPER_PROPERTY,
                        length: node.arguments.length
                    }, node.base, node.type);
                }
                else {
                    yield this.process(base);
                    this.context.pushCode({
                        op: instruction_1.OpCode.PUSH,
                        value: new string_1.CustomString(left.identifier.name)
                    }, left.identifier);
                    yield pushArgs();
                    this.context.pushCode({
                        op: instruction_1.OpCode.CALL_WITH_CONTEXT,
                        length: node.arguments.length
                    }, left.identifier, node.type);
                }
            }
            else if (left instanceof miniscript_core_1.ASTIndexExpression) {
                const base = (0, utils_1.unwrap)(left.base);
                if (base instanceof miniscript_core_1.ASTIdentifier && base.name === keywords_1.RuntimeKeyword.Super) {
                    yield this.process(left.index);
                    yield pushArgs();
                    this.context.pushCode({
                        op: instruction_1.OpCode.CALL_SUPER_PROPERTY,
                        length: node.arguments.length
                    }, left.index, node.type);
                }
                else {
                    yield this.process(base);
                    yield this.process(left.index);
                    yield pushArgs();
                    this.context.pushCode({
                        op: instruction_1.OpCode.CALL_WITH_CONTEXT,
                        length: node.arguments.length
                    }, left.index, node.type);
                }
            }
            else if (left instanceof miniscript_core_1.ASTIdentifier) {
                yield this.processIdentifier(left, {
                    isDescending: false,
                    isReference: true
                });
                yield pushArgs();
                this.context.pushCode({
                    op: instruction_1.OpCode.CALL,
                    length: node.arguments.length
                }, left, node.type);
            }
            else {
                yield this.process(left);
                yield pushArgs();
                this.context.pushCode({
                    op: instruction_1.OpCode.CALL,
                    length: node.arguments.length
                }, left, node.type);
            }
        });
    }
    processEnvarExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            this.context.pushCode({
                op: instruction_1.OpCode.PUSH,
                value: new string_1.CustomString(node.name)
            }, node);
            this.context.pushCode({
                op: instruction_1.OpCode.GET_ENVAR
            }, node);
        });
    }
}
exports.BytecodeExpressionGenerator = BytecodeExpressionGenerator;
