"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BytecodeStatementGenerator = void 0;
const greybel_core_1 = require("greybel-core");
const miniscript_core_1 = require("miniscript-core");
const default_1 = require("../types/default");
const number_1 = require("../types/number");
const string_1 = require("../types/string");
const error_1 = require("../utils/error");
const expression_1 = require("./expression");
const instruction_1 = require("./instruction");
const keywords_1 = require("./keywords");
const module_1 = require("./module");
const utils_1 = require("./utils");
class BytecodeStatementGenerator {
    constructor(context, parseCodeFunction) {
        this.context = context;
        this.exprGenerator = new expression_1.BytecodeExpressionGenerator(this.context, parseCodeFunction, this);
        this.parseCode = parseCodeFunction;
    }
    process(node) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.context.isDebugMode()) {
                this.context.pushCode({
                    op: instruction_1.OpCode.BREAKPOINT
                }, node);
            }
            switch (node.type) {
                case miniscript_core_1.ASTType.MemberExpression:
                    yield this.processMemberExpression(node);
                    return;
                case miniscript_core_1.ASTType.IndexExpression:
                    yield this.processIndexExpression(node);
                    return;
                case miniscript_core_1.ASTType.SliceExpression:
                    return;
                case miniscript_core_1.ASTType.Identifier:
                    yield this.processIdentifier(node);
                    return;
                case miniscript_core_1.ASTType.AssignmentStatement:
                    yield this.processAssignmentStatement(node);
                    return;
                case miniscript_core_1.ASTType.Chunk: {
                    const chunk = node;
                    for (const item of chunk.body) {
                        yield this.process(item);
                    }
                    return;
                }
                case miniscript_core_1.ASTType.BooleanLiteral:
                case miniscript_core_1.ASTType.StringLiteral:
                case miniscript_core_1.ASTType.NumericLiteral:
                case miniscript_core_1.ASTType.NilLiteral:
                    return;
                case miniscript_core_1.ASTType.IsaExpression:
                case miniscript_core_1.ASTType.BinaryExpression:
                case miniscript_core_1.ASTType.LogicalExpression:
                    yield this.processEvaluationExpression(node);
                    return;
                case miniscript_core_1.ASTType.ReturnStatement:
                    yield this.processReturn(node);
                    return;
                case miniscript_core_1.ASTType.BreakStatement:
                    yield this.processBreak(node);
                    return;
                case miniscript_core_1.ASTType.ContinueStatement:
                    yield this.processContinue(node);
                    return;
                case miniscript_core_1.ASTType.MapConstructorExpression:
                    yield this.processMapConstructorExpression(node);
                    return;
                case miniscript_core_1.ASTType.ListConstructorExpression:
                    yield this.processListConstructorExpression(node);
                    return;
                case miniscript_core_1.ASTType.FunctionDeclaration:
                    return;
                case miniscript_core_1.ASTType.WhileStatement:
                    yield this.processWhileStatement(node);
                    return;
                case miniscript_core_1.ASTType.ParenthesisExpression:
                    yield this.process(node.expression);
                    return;
                case miniscript_core_1.ASTType.BinaryNegatedExpression:
                case miniscript_core_1.ASTType.UnaryExpression:
                case miniscript_core_1.ASTType.NegationExpression:
                    yield this.processUnaryExpression(node);
                    return;
                case miniscript_core_1.ASTType.CallStatement:
                    yield this.process(node.expression);
                    return;
                case miniscript_core_1.ASTType.CallExpression:
                    yield this.processCallExpression(node);
                    return;
                case miniscript_core_1.ASTType.IfStatement:
                case miniscript_core_1.ASTType.IfShortcutStatement:
                    yield this.processIfStatement(node);
                    return;
                case miniscript_core_1.ASTType.ForGenericStatement:
                    yield this.processForGenericStatement(node);
                    return;
                case miniscript_core_1.ASTType.EmptyExpression:
                case miniscript_core_1.ASTType.Comment:
                    return;
                case greybel_core_1.ASTType.FeatureEnvarExpression:
                    return;
                case greybel_core_1.ASTType.FeatureImportExpression:
                    yield this.processImportExpression(node);
                    return;
                case greybel_core_1.ASTType.FeatureIncludeExpression:
                    yield this.processIncludeExpression(node);
                    return;
                case greybel_core_1.ASTType.FeatureDebuggerExpression:
                    yield this.processDebuggerExpression(node);
                    return;
                default: {
                    const range = new miniscript_core_1.ASTRange(node.start, node.end);
                    throw new error_1.PrepareError(`Unexpected AST type ${node.type}`, {
                        target: this.context.target.peek(),
                        range
                    });
                }
            }
        });
    }
    processMemberExpression(node, context) {
        return __awaiter(this, void 0, void 0, function* () {
            const base = (0, utils_1.unwrap)(node.base);
            if (base instanceof miniscript_core_1.ASTIdentifier && base.name === keywords_1.RuntimeKeyword.Super) {
                this.context.pushCode({
                    op: instruction_1.OpCode.PUSH,
                    value: new string_1.CustomString(node.identifier.name)
                }, node.identifier);
                this.context.pushCode({
                    op: instruction_1.OpCode.GET_SUPER_PROPERTY,
                    invoke: !(context === null || context === void 0 ? void 0 : context.isReference),
                    command: true
                }, node.identifier);
            }
            else {
                yield this.exprGenerator.process(base);
                yield this.processIdentifier(node.identifier, {
                    isDescending: true,
                    isReference: !!(context === null || context === void 0 ? void 0 : context.isReference)
                });
            }
        });
    }
    processIndexExpression(node, context) {
        return __awaiter(this, void 0, void 0, function* () {
            const base = (0, utils_1.unwrap)(node.base);
            if (base instanceof miniscript_core_1.ASTIdentifier && base.name === keywords_1.RuntimeKeyword.Super) {
                yield this.exprGenerator.process(node.index);
                this.context.pushCode({
                    op: instruction_1.OpCode.GET_SUPER_PROPERTY,
                    invoke: !(context === null || context === void 0 ? void 0 : context.isReference),
                    command: true
                }, node.index, node.type);
            }
            else {
                yield this.exprGenerator.process(base);
                yield this.exprGenerator.process(node.index);
                this.context.pushCode({
                    op: instruction_1.OpCode.GET_PROPERTY,
                    invoke: !(context === null || context === void 0 ? void 0 : context.isReference),
                    command: true
                }, node.index, node.type);
            }
        });
    }
    processIdentifier(node, context) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!(context === null || context === void 0 ? void 0 : context.isDescending)) {
                switch (node.name) {
                    case keywords_1.RuntimeKeyword.Self:
                    case keywords_1.RuntimeKeyword.Super:
                    case keywords_1.RuntimeKeyword.Outer:
                    case keywords_1.RuntimeKeyword.Locals:
                    case keywords_1.RuntimeKeyword.Globals: {
                        return;
                    }
                    default: {
                        this.context.pushCode({
                            op: instruction_1.OpCode.GET_VARIABLE,
                            property: new string_1.CustomString(node.name),
                            invoke: !(context === null || context === void 0 ? void 0 : context.isReference),
                            command: true
                        }, node);
                    }
                }
            }
            else {
                this.context.pushCode({
                    op: instruction_1.OpCode.PUSH,
                    value: new string_1.CustomString(node.name)
                }, node);
                this.context.pushCode({
                    op: instruction_1.OpCode.GET_PROPERTY,
                    invoke: !(context === null || context === void 0 ? void 0 : context.isReference),
                    command: true
                }, node);
            }
        });
    }
    processAssignmentStatement(node) {
        return __awaiter(this, void 0, void 0, function* () {
            let variable = (0, utils_1.unwrap)(node.variable);
            if (variable instanceof miniscript_core_1.ASTUnaryExpression) {
                variable = (0, utils_1.unwrap)(variable.argument);
            }
            if (variable instanceof miniscript_core_1.ASTMemberExpression) {
                yield this.exprGenerator.process(variable.base);
                this.context.pushCode({
                    op: instruction_1.OpCode.PUSH,
                    value: new string_1.CustomString(variable.identifier.name)
                }, variable.identifier);
            }
            else if (variable instanceof miniscript_core_1.ASTIndexExpression) {
                yield this.exprGenerator.process(variable.base);
                yield this.exprGenerator.process(variable.index);
            }
            else if (variable instanceof miniscript_core_1.ASTIdentifier) {
                this.context.pushCode({
                    op: instruction_1.OpCode.GET_LOCALS
                }, variable);
                this.context.pushCode({
                    op: instruction_1.OpCode.PUSH,
                    value: new string_1.CustomString(variable.name)
                }, variable);
            }
            else {
                yield this.exprGenerator.process(variable);
                this.context.pushCode({
                    op: instruction_1.OpCode.PUSH,
                    value: default_1.DefaultType.Void
                }, variable);
            }
            yield this.exprGenerator.process(node.init, { includeOuter: true });
            this.context.pushCode({
                op: instruction_1.OpCode.ASSIGN
            }, node);
        });
    }
    processEvaluationExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const skip = {
                op: instruction_1.OpCode.NOOP
            };
            yield this.exprGenerator.process(node.left);
            if (node.operator === miniscript_core_1.Operator.And) {
                this.context.pushCode({
                    op: instruction_1.OpCode.GOTO_A_IF_FALSE,
                    goto: skip
                }, node);
            }
            else if (node.operator === miniscript_core_1.Operator.Or) {
                this.context.pushCode({
                    op: instruction_1.OpCode.GOTO_A_IF_TRUE,
                    goto: skip
                }, node);
            }
            yield this.process(node.right);
            switch (node.operator) {
                case miniscript_core_1.Operator.And:
                case miniscript_core_1.Operator.Or: {
                    this.context.pushCode(skip, node);
                    break;
                }
                case miniscript_core_1.Operator.Isa:
                case miniscript_core_1.Operator.Equal:
                case miniscript_core_1.Operator.NotEqual:
                case miniscript_core_1.Operator.LessThan:
                case miniscript_core_1.Operator.LessThanOrEqual:
                case miniscript_core_1.Operator.GreaterThan:
                case miniscript_core_1.Operator.GreaterThanOrEqual:
                case miniscript_core_1.Operator.Plus:
                case miniscript_core_1.Operator.Minus:
                case miniscript_core_1.Operator.Asterik:
                case miniscript_core_1.Operator.Slash:
                case miniscript_core_1.Operator.Modulo:
                case miniscript_core_1.Operator.Power:
                case greybel_core_1.Operator.BitwiseAnd:
                case greybel_core_1.Operator.BitwiseOr:
                case greybel_core_1.Operator.LeftShift:
                case greybel_core_1.Operator.RightShift:
                case greybel_core_1.Operator.UnsignedRightShift: {
                    break;
                }
                default:
                    throw new Error(`Unexpected evaluation expression operator. ("${node.operator}")`);
            }
        });
    }
    processReturn(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const mod = this.context.module.peek();
            if (mod.isGlobalScope()) {
                if (node.argument) {
                    yield this.process(node.argument);
                }
                return;
            }
            if (node.argument) {
                yield this.exprGenerator.process(node.argument);
            }
            else {
                this.context.pushCode({
                    op: instruction_1.OpCode.PUSH,
                    value: default_1.DefaultType.Void
                }, node);
            }
            this.context.pushCode({
                op: instruction_1.OpCode.RETURN
            }, node);
        });
    }
    processBreak(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const mod = this.context.module.peek();
            const jumpPoint = mod.getJumpPoint();
            if (jumpPoint === null)
                return;
            const end = jumpPoint[1];
            this.context.pushCode({
                op: instruction_1.OpCode.GOTO_A,
                goto: end
            }, node);
        });
    }
    processContinue(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const mod = this.context.module.peek();
            const jumpPoint = mod.getJumpPoint();
            if (jumpPoint === null)
                return;
            const start = jumpPoint[0];
            this.context.pushCode({
                op: instruction_1.OpCode.GOTO_A,
                goto: start
            }, node);
        });
    }
    processMapConstructorExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const field of node.fields) {
                yield this.exprGenerator.process(field.key);
                yield this.exprGenerator.process(field.value);
            }
            this.context.pushCode({
                op: instruction_1.OpCode.CONSTRUCT_MAP,
                length: node.fields.length,
                command: true
            }, node);
        });
    }
    processListConstructorExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const field of node.fields) {
                yield this.exprGenerator.process(field.value);
            }
            this.context.pushCode({
                op: instruction_1.OpCode.CONSTRUCT_LIST,
                length: node.fields.length,
                command: true
            }, node);
        });
    }
    processWhileStatement(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const mod = this.context.module.peek();
            const start = {
                op: instruction_1.OpCode.NOOP
            };
            const end = {
                op: instruction_1.OpCode.NOOP
            };
            this.context.pushCode(start, node.condition);
            yield this.exprGenerator.process(node.condition);
            this.context.pushCode({
                op: instruction_1.OpCode.GOTO_A_IF_FALSE,
                goto: end
            }, node.condition);
            mod.pushJumppoint(start, end);
            for (const item of node.body) {
                yield this.process(item);
            }
            mod.popJumppoint();
            this.context.pushCode({
                op: instruction_1.OpCode.GOTO_A,
                goto: start
            }, node.condition);
            this.context.pushCode(end, node);
        });
    }
    processUnaryExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const arg = (0, utils_1.unwrap)(node.argument);
            switch (node.operator) {
                case miniscript_core_1.Operator.Reference:
                    if (arg instanceof miniscript_core_1.ASTMemberExpression) {
                        yield this.processMemberExpression(arg, { isReference: true });
                    }
                    else if (arg instanceof miniscript_core_1.ASTIndexExpression) {
                        yield this.processIndexExpression(arg, { isReference: true });
                    }
                    else if (arg instanceof miniscript_core_1.ASTIdentifier) {
                        yield this.processIdentifier(arg, {
                            isDescending: false,
                            isReference: true
                        });
                    }
                    else {
                        yield this.process(arg);
                    }
                    return;
                case miniscript_core_1.Operator.Not: {
                    yield this.exprGenerator.process(arg);
                    this.context.pushCode({
                        op: instruction_1.OpCode.FALSIFY,
                        command: true
                    }, node);
                    return;
                }
                case miniscript_core_1.Operator.Minus: {
                    yield this.exprGenerator.process(arg);
                    this.context.pushCode({
                        op: instruction_1.OpCode.NEGATE,
                        command: true
                    }, node);
                    return;
                }
                case miniscript_core_1.Operator.New: {
                    yield this.exprGenerator.process(arg);
                    this.context.pushCode({
                        op: instruction_1.OpCode.NEW,
                        command: true
                    }, node);
                }
            }
        });
    }
    processCallExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const pushArgs = () => __awaiter(this, void 0, void 0, function* () {
                for (const arg of node.arguments) {
                    yield this.exprGenerator.process(arg);
                }
            });
            const left = (0, utils_1.unwrap)(node.base);
            if (left instanceof miniscript_core_1.ASTMemberExpression) {
                const base = (0, utils_1.unwrap)(left.base);
                if (base instanceof miniscript_core_1.ASTIdentifier && base.name === keywords_1.RuntimeKeyword.Super) {
                    this.context.pushCode({
                        op: instruction_1.OpCode.PUSH,
                        value: new string_1.CustomString(left.identifier.name)
                    }, left.identifier);
                    yield pushArgs();
                    this.context.pushCode({
                        op: instruction_1.OpCode.CALL_SUPER_PROPERTY,
                        length: node.arguments.length,
                        command: true
                    }, node.base, node.type);
                }
                else {
                    yield this.exprGenerator.process(base);
                    this.context.pushCode({
                        op: instruction_1.OpCode.PUSH,
                        value: new string_1.CustomString(left.identifier.name)
                    }, left.identifier);
                    yield pushArgs();
                    this.context.pushCode({
                        op: instruction_1.OpCode.CALL_WITH_CONTEXT,
                        length: node.arguments.length,
                        command: true
                    }, left.identifier, node.type);
                }
            }
            else if (left instanceof miniscript_core_1.ASTIndexExpression) {
                const base = (0, utils_1.unwrap)(left.base);
                if (base instanceof miniscript_core_1.ASTIdentifier && base.name === keywords_1.RuntimeKeyword.Super) {
                    yield this.exprGenerator.process(left.index);
                    yield pushArgs();
                    this.context.pushCode({
                        op: instruction_1.OpCode.CALL_SUPER_PROPERTY,
                        length: node.arguments.length,
                        command: true
                    }, left.index, node.type);
                }
                else {
                    yield this.exprGenerator.process(base);
                    yield this.exprGenerator.process(left.index);
                    yield pushArgs();
                    this.context.pushCode({
                        op: instruction_1.OpCode.CALL_WITH_CONTEXT,
                        length: node.arguments.length,
                        command: true
                    }, left.index, node.type);
                }
            }
            else if (left instanceof miniscript_core_1.ASTIdentifier) {
                yield this.exprGenerator.processIdentifier(left, {
                    isDescending: false,
                    isReference: true
                });
                yield pushArgs();
                this.context.pushCode({
                    op: instruction_1.OpCode.CALL,
                    length: node.arguments.length,
                    command: true
                }, left, node.type);
            }
            else {
                yield this.exprGenerator.process(left);
                yield pushArgs();
                this.context.pushCode({
                    op: instruction_1.OpCode.CALL,
                    length: node.arguments.length,
                    command: true
                }, left, node.type);
            }
        });
    }
    processIfStatement(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const end = {
                op: instruction_1.OpCode.NOOP
            };
            for (const clause of node.clauses) {
                if (clause instanceof miniscript_core_1.ASTIfClause) {
                    const next = {
                        op: instruction_1.OpCode.NOOP
                    };
                    yield this.exprGenerator.process(clause.condition);
                    this.context.pushCode({
                        op: instruction_1.OpCode.GOTO_A_IF_FALSE,
                        goto: next
                    }, node);
                    for (const item of clause.body) {
                        yield this.process(item);
                    }
                    this.context.pushCode({
                        op: instruction_1.OpCode.GOTO_A,
                        goto: end
                    }, node);
                    this.context.pushCode(next, clause);
                }
                else if (clause instanceof miniscript_core_1.ASTElseClause) {
                    for (const item of clause.body) {
                        yield this.process(item);
                    }
                }
            }
            this.context.pushCode(end, node);
        });
    }
    processForGenericStatement(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const mod = this.context.module.peek();
            const variable = node.variable;
            const idxVariable = new string_1.CustomString(`__${variable.name}_idx`);
            const start = {
                op: instruction_1.OpCode.NEXT,
                idxVariable,
                variable: new string_1.CustomString(variable.name)
            };
            const end = {
                op: instruction_1.OpCode.POP_ITERATOR
            };
            mod.pushJumppoint(start, end);
            this.context.pushCode({
                op: instruction_1.OpCode.GET_LOCALS
            }, node);
            this.context.pushCode({
                op: instruction_1.OpCode.PUSH,
                value: idxVariable
            }, node);
            this.context.pushCode({
                op: instruction_1.OpCode.PUSH,
                value: new number_1.CustomNumber(-1)
            }, node);
            this.context.pushCode({
                op: instruction_1.OpCode.ASSIGN
            }, node);
            yield this.exprGenerator.process(node.iterator);
            this.context.pushCode({
                op: instruction_1.OpCode.PUSH_ITERATOR
            }, node.iterator);
            this.context.pushCode(start, node.iterator);
            this.context.pushCode({
                op: instruction_1.OpCode.GOTO_A_IF_FALSE,
                goto: end
            }, node.iterator);
            for (const item of node.body) {
                yield this.process(item);
            }
            this.context.pushCode({
                op: instruction_1.OpCode.GOTO_A,
                goto: start
            }, node.iterator);
            this.context.pushCode(end, node.iterator);
            mod.popJumppoint();
        });
    }
    createImport(node, path, code) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const mod = new module_1.Module(path);
                this.context.target.push(path);
                this.context.module.push(mod);
                const childNodes = this.parseCode(code);
                this.context.pushInternalCode({
                    op: instruction_1.OpCode.GET_LOCALS
                });
                this.context.pushInternalCode({
                    op: instruction_1.OpCode.PUSH,
                    value: new string_1.CustomString('module')
                });
                this.context.pushInternalCode({
                    op: instruction_1.OpCode.PUSH,
                    value: new string_1.CustomString('exports')
                });
                this.context.pushInternalCode({
                    op: instruction_1.OpCode.CONSTRUCT_MAP,
                    length: 0
                });
                this.context.pushInternalCode({
                    op: instruction_1.OpCode.CONSTRUCT_MAP,
                    length: 1
                });
                this.context.pushInternalCode({
                    op: instruction_1.OpCode.ASSIGN
                });
                yield this.process(childNodes);
                this.context.pushInternalCode({
                    op: instruction_1.OpCode.GET_VARIABLE,
                    property: new string_1.CustomString('module')
                });
                this.context.pushInternalCode({
                    op: instruction_1.OpCode.PUSH,
                    value: new string_1.CustomString('exports')
                });
                this.context.pushInternalCode({
                    op: instruction_1.OpCode.GET_PROPERTY
                });
                this.context.module.pop();
                this.context.target.pop();
                this.context.imports.set(path, mod.getCode());
            }
            catch (err) {
                if (err instanceof error_1.PrepareError) {
                    throw err;
                }
                throw new error_1.PrepareError(err.message, {
                    target: path,
                    range: new miniscript_core_1.ASTRange(node.start, node.end)
                }, err);
            }
        });
    }
    processImportExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const currentTarget = this.context.target.peek();
            const importTarget = yield this.context.handler.resourceHandler.getTargetRelativeTo(currentTarget, node.path);
            if (this.context.target.includes(importTarget)) {
                console.warn(`Found circular dependency between "${currentTarget}" and "${importTarget}" at line ${node.start.line}. Using noop instead to prevent overflow.`);
                return;
            }
            if (!this.context.imports.has(importTarget)) {
                const code = yield this.context.handler.resourceHandler.get(importTarget);
                if (code == null) {
                    const range = new miniscript_core_1.ASTRange(node.start, node.end);
                    throw new error_1.PrepareError(`Cannot find import "${currentTarget}"`, {
                        target: currentTarget,
                        range
                    });
                }
                yield this.createImport(node, importTarget, code);
            }
            this.context.pushCode({
                op: instruction_1.OpCode.GET_LOCALS
            }, node);
            this.context.pushCode({
                op: instruction_1.OpCode.PUSH,
                value: new string_1.CustomString(node.name.name)
            }, node);
            this.context.pushCode({
                op: instruction_1.OpCode.IMPORT,
                path: importTarget
            }, node);
            this.context.pushCode({
                op: instruction_1.OpCode.ASSIGN
            }, node);
        });
    }
    processIncludeExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            const currentTarget = this.context.target.peek();
            const importTarget = yield this.context.handler.resourceHandler.getTargetRelativeTo(currentTarget, node.path);
            if (this.context.target.includes(importTarget)) {
                console.warn(`Found circular dependency between "${currentTarget}" and "${importTarget}" at line ${node.start.line}. Using noop instead to prevent overflow.`);
                return;
            }
            const code = yield this.context.handler.resourceHandler.get(importTarget);
            if (code == null) {
                const range = new miniscript_core_1.ASTRange(node.start, node.end);
                throw new error_1.PrepareError(`Cannot find import "${currentTarget}"`, {
                    target: currentTarget,
                    range
                });
            }
            try {
                this.context.target.push(importTarget);
                const childNodes = this.parseCode(code);
                yield this.process(childNodes);
                this.context.target.pop();
            }
            catch (err) {
                if (err instanceof error_1.PrepareError) {
                    throw err;
                }
                throw new error_1.PrepareError(err.message, {
                    target: importTarget,
                    range: new miniscript_core_1.ASTRange(node.start, node.end)
                }, err);
            }
        });
    }
    processDebuggerExpression(node) {
        return __awaiter(this, void 0, void 0, function* () {
            this.context.pushCode({
                op: instruction_1.OpCode.BREAKPOINT_ENABLE
            }, node);
            this.context.pushCode({
                op: instruction_1.OpCode.BREAKPOINT
            }, node);
        });
    }
}
exports.BytecodeStatementGenerator = BytecodeStatementGenerator;
